@@ -43,6 +43,7 @@ struct connman_device {
 	connman_uint16_t scan_interval;
 	char *name;
 	char *node;
+	char *address;
 	char *interface;
 	unsigned int connections;
 	guint scan_timeout;
@@ -88,6 +89,8 @@ static const char *type2description(enum connman_device_type type)
 		return "WiMAX";
 	case CONNMAN_DEVICE_TYPE_BLUETOOTH:
 		return "Bluetooth";
+	case CONNMAN_DEVICE_TYPE_GPS:
+		return "GPS";
 	case CONNMAN_DEVICE_TYPE_HSO:
 	case CONNMAN_DEVICE_TYPE_NOZOMI:
 	case CONNMAN_DEVICE_TYPE_HUAWEI:
@@ -112,6 +115,8 @@ static const char *type2string(enum connman_device_type type)
 		return "wimax";
 	case CONNMAN_DEVICE_TYPE_BLUETOOTH:
 		return "bluetooth";
+	case CONNMAN_DEVICE_TYPE_GPS:
+		return "gps";
 	case CONNMAN_DEVICE_TYPE_HSO:
 	case CONNMAN_DEVICE_TYPE_HUAWEI:
 	case CONNMAN_DEVICE_TYPE_NOZOMI:
@@ -165,16 +170,18 @@ static int set_powered(struct connman_device *device, connman_bool_t powered)
 		return -EINVAL;
 
 	if (powered == TRUE) {
-		if (driver->enable)
+		if (driver->enable) {
 			err = driver->enable(device);
-		else
+			__connman_notifier_device_type_increase(device->type);
+		} else
 			err = -EINVAL;
 	} else {
 		g_hash_table_remove_all(device->networks);
 
-		if (driver->disable)
+		if (driver->disable) {
 			err = driver->disable(device);
-		else
+			__connman_notifier_device_type_decrease(device->type);
+		} else
 			err = -EINVAL;
 	}
 
@@ -305,6 +312,10 @@ static DBusMessage *get_properties(DBusConnection *conn,
 		connman_dbus_dict_append_variant(&dict, "Type",
 						DBUS_TYPE_STRING, &str);
 
+	if (device->address != NULL)
+		connman_dbus_dict_append_variant(&dict, "Address",
+					DBUS_TYPE_STRING, &device->address);
+
 	if (device->interface != NULL)
 		connman_dbus_dict_append_variant(&dict, "Interface",
 					DBUS_TYPE_STRING, &device->interface);
@@ -459,6 +470,78 @@ static DBusMessage *set_property(DBusConnection *conn,
 	return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
 }
 
+static DBusMessage *join_network(DBusConnection *conn,
+					DBusMessage *msg, void *data)
+{
+	struct connman_device *device = data;
+	struct connman_network *network;
+	enum connman_network_type type;
+	DBusMessageIter iter, array;
+	int err, index;
+
+	DBG("conn %p", conn);
+
+	if (__connman_security_check_privilege(msg,
+					CONNMAN_SECURITY_PRIVILEGE_MODIFY) < 0)
+		return __connman_error_permission_denied(msg);
+
+	if (!device->driver || !device->driver->join)
+		return __connman_error_not_supported(msg);
+
+	dbus_message_iter_init(msg, &iter);
+	dbus_message_iter_recurse(&iter, &array);
+
+	switch (device->type) {
+	case CONNMAN_DEVICE_TYPE_WIFI:
+		type = CONNMAN_NETWORK_TYPE_WIFI;
+		break;
+	default:
+		return __connman_error_not_supported(msg);
+	}
+
+	network = connman_network_create("00_00_00_00_00_00", type);
+	if (network == NULL)
+		return __connman_error_failed(msg);
+
+	while (dbus_message_iter_get_arg_type(&array) == DBUS_TYPE_DICT_ENTRY) {
+		DBusMessageIter entry, value;
+		const char *key, *str;
+
+		dbus_message_iter_recurse(&array, &entry);
+		dbus_message_iter_get_basic(&entry, &key);
+
+		dbus_message_iter_next(&entry);
+		dbus_message_iter_recurse(&entry, &value);
+
+		switch (dbus_message_iter_get_arg_type(&value)) {
+		case DBUS_TYPE_STRING:
+			dbus_message_iter_get_basic(&value, &str);
+			if (g_str_equal(key, "WiFi.SSID") == TRUE)
+				connman_network_set_blob(network, key,
+							str, strlen(str));
+			else
+				connman_network_set_string(network, key, str);
+			break;
+		}
+
+		dbus_message_iter_next(&array);
+	}
+
+	index = connman_device_get_index(device);
+	connman_network_set_index(network, index);
+
+	connman_network_set_protocol(network, CONNMAN_NETWORK_PROTOCOL_IP);
+
+	err = device->driver->join(device, network);
+
+	connman_network_unref(network);
+
+	if (err < 0)
+		return __connman_error_failed(msg);
+
+	return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
+}
+
 static DBusMessage *create_network(DBusConnection *conn,
 					DBusMessage *msg, void *data)
 {
@@ -516,6 +599,7 @@ static DBusMessage *propose_scan(DBusConnection *conn,
 static GDBusMethodTable device_methods[] = {
 	{ "GetProperties", "",      "a{sv}", get_properties },
 	{ "SetProperty",   "sv",    "",      set_property   },
+	{ "JoinNetwork",   "a{sv}", "",      join_network   },
 	{ "CreateNetwork", "a{sv}", "o",     create_network },
 	{ "RemoveNetwork", "o",     "",      remove_network },
 	{ "ProposeScan",   "",      "",      propose_scan   },
@@ -611,8 +695,10 @@ static void device_enable(struct connman_device *device)
 	if (device->powered == TRUE)
 		return;
 
-	if (device->driver->enable)
+	if (device->driver->enable) {
 		device->driver->enable(device);
+		__connman_notifier_device_type_increase(device->type);
+	}
 }
 
 static void device_disable(struct connman_device *device)
@@ -627,8 +713,10 @@ static void device_disable(struct connman_device *device)
 
 	g_hash_table_remove_all(device->networks);
 
-	if (device->driver->disable)
+	if (device->driver->disable) {
 		device->driver->disable(device);
+		__connman_notifier_device_type_decrease(device->type);
+	}
 }
 
 static int setup_device(struct connman_device *device)
@@ -645,6 +733,16 @@ static int setup_device(struct connman_device *device)
 		return err;
 	}
 
+	switch (device->mode) {
+	case CONNMAN_DEVICE_MODE_UNKNOWN:
+	case CONNMAN_DEVICE_MODE_NETWORK_SINGLE:
+	case CONNMAN_DEVICE_MODE_NETWORK_MULTIPLE:
+		break;
+	case CONNMAN_DEVICE_MODE_TRANSPORT_IP:
+		__connman_profile_add_device(device);
+		break;
+	}
+
 	device_enable(device);
 
 	return 0;
@@ -679,6 +777,16 @@ static void remove_device(struct connman_device *device)
 
 	device_disable(device);
 
+	switch (device->mode) {
+	case CONNMAN_DEVICE_MODE_UNKNOWN:
+	case CONNMAN_DEVICE_MODE_NETWORK_SINGLE:
+	case CONNMAN_DEVICE_MODE_NETWORK_MULTIPLE:
+		break;
+	case CONNMAN_DEVICE_MODE_TRANSPORT_IP:
+		__connman_profile_remove_device(device);
+		break;
+	}
+
 	unregister_interface(&device->element);
 
 	if (device->driver->remove)
@@ -774,6 +882,7 @@ static void device_destruct(struct connman_element *element)
 
 	g_free(device->node);
 	g_free(device->name);
+	g_free(device->address);
 	g_free(device->interface);
 
 	g_free(device->last_network);
@@ -815,8 +924,8 @@ struct connman_device *connman_device_create(const char *node,
 
 	str = type2string(type);
 	if (str != NULL)
-		connman_element_set_static_property(&device->element,
-					"Type", DBUS_TYPE_STRING, &str);
+		connman_element_set_string(&device->element,
+					CONNMAN_PROPERTY_ID_TYPE, str);
 
 	device->element.ipv4.method = CONNMAN_IPV4_METHOD_DHCP;
 
@@ -843,6 +952,10 @@ struct connman_device *connman_device_create(const char *node,
 		device->priority = 50;
 		device->scan_interval = 0;
 		break;
+	case CONNMAN_DEVICE_TYPE_GPS:
+		device->priority = 0;
+		device->scan_interval = 0;
+		break;
 	case CONNMAN_DEVICE_TYPE_HSO:
 	case CONNMAN_DEVICE_TYPE_NOZOMI:
 	case CONNMAN_DEVICE_TYPE_HUAWEI:
@@ -1153,7 +1266,8 @@ static void connect_known_network(struct connman_device *device)
 		if (connman_network_get_available(value) == FALSE)
 			continue;
 
-		name = connman_network_get_string(value, "Name");
+		name = connman_network_get_string(value,
+						CONNMAN_PROPERTY_ID_NAME);
 		if (name != NULL && device->last_network != NULL) {
 			if (g_str_equal(name, device->last_network) == TRUE) {
 				network = value;
@@ -1169,8 +1283,10 @@ static void connect_known_network(struct connman_device *device)
 			continue;
 		}
 
-		old_priority = connman_network_get_uint8(network, "Priority");
-		new_priority = connman_network_get_uint8(value, "Priority");
+		old_priority = connman_network_get_uint8(network,
+						CONNMAN_PROPERTY_ID_PRIORITY);
+		new_priority = connman_network_get_uint8(value,
+						CONNMAN_PROPERTY_ID_PRIORITY);
 
 		if (new_priority != old_priority) {
 			if (new_priority > old_priority)
@@ -1178,8 +1294,10 @@ static void connect_known_network(struct connman_device *device)
 			continue;
 		}
 
-		old_strength = connman_network_get_uint8(network, "Strength");
-		new_strength = connman_network_get_uint8(value, "Strength");
+		old_strength = connman_network_get_uint8(network,
+						CONNMAN_PROPERTY_ID_STRENGTH);
+		new_strength = connman_network_get_uint8(value,
+						CONNMAN_PROPERTY_ID_STRENGTH);
 
 		if (new_strength > old_strength)
 			network = value;
@@ -1344,7 +1462,10 @@ int connman_device_set_string(struct connman_device *device,
 {
 	DBG("device %p key %s value %s", device, key, value);
 
-	if (g_str_equal(key, "Name") == TRUE) {
+	if (g_str_equal(key, "Address") == TRUE) {
+		g_free(device->address);
+		device->address = g_strdup(value);
+	} else if (g_str_equal(key, "Name") == TRUE) {
 		g_free(device->name);
 		device->name = g_strdup(value);
 	} else if (g_str_equal(key, "Node") == TRUE) {
@@ -1352,7 +1473,7 @@ int connman_device_set_string(struct connman_device *device,
 		device->node = g_strdup(value);
 	}
 
-	return 0;
+	return connman_element_set_string(&device->element, key, value);
 }
 
 /**
@@ -1367,12 +1488,14 @@ const char *connman_device_get_string(struct connman_device *device,
 {
 	DBG("device %p key %s", device, key);
 
-	if (g_str_equal(key, "Name") == TRUE)
+	if (g_str_equal(key, "Address") == TRUE)
+		return device->address;
+	else if (g_str_equal(key, "Name") == TRUE)
 		return device->name;
 	else if (g_str_equal(key, "Node") == TRUE)
 		return device->node;
 
-	return NULL;
+	return connman_element_get_string(&device->element, key);
 }
 
 static void set_offlinemode(struct connman_element *element, gpointer user_data)
@@ -1401,6 +1524,8 @@ int __connman_device_set_offlinemode(connman_bool_t offlinemode)
 	__connman_element_foreach(NULL, CONNMAN_ELEMENT_TYPE_DEVICE,
 			set_offlinemode, GUINT_TO_POINTER(offlinemode));
 
+	__connman_notifier_offline_mode(offlinemode);
+
 	return 0;
 }
 
@@ -1493,7 +1618,8 @@ void __connman_device_set_network(struct connman_device *device,
 	const char *name;
 
 	if (network != NULL) {
-		name = connman_network_get_string(network, "Name");
+		name = connman_network_get_string(network,
+						CONNMAN_PROPERTY_ID_NAME);
 		device->last_network = g_strdup(name);
 	}
 
@@ -1681,10 +1807,12 @@ static int device_load(struct connman_device *device)
 		break;
 	}
 
+#if 0
 	str = g_key_file_get_string(keyfile, "Configuration",
 							"LastNetwork", NULL);
 	if (str != NULL)
 		device->last_network = str;
+#endif
 
 	g_key_file_free(keyfile);
 
