@@ -39,6 +39,7 @@ struct connman_network {
 	connman_uint8_t priority;
 	connman_uint8_t strength;
 	char *identifier;
+	char *address;
 	char *name;
 	char *node;
 
@@ -109,6 +110,10 @@ static DBusMessage *get_properties(DBusConnection *conn,
 						DBUS_TYPE_OBJECT_PATH, &path);
 	}
 
+	if (network->address != NULL)
+		connman_dbus_dict_append_variant(&dict, "Address",
+					DBUS_TYPE_STRING, &network->address);
+
 	if (network->name != NULL)
 		connman_dbus_dict_append_variant(&dict, "Name",
 					DBUS_TYPE_STRING, &network->name);
@@ -430,6 +435,7 @@ static void network_destruct(struct connman_element *element)
 
 	g_free(network->node);
 	g_free(network->name);
+	g_free(network->address);
 	g_free(network->identifier);
 }
 
@@ -815,6 +821,8 @@ int __connman_network_disconnect(struct connman_network *network)
 	if (network->connected == FALSE)
 		return -ENOTCONN;
 
+	connman_element_unregister_children(&network->element);
+
 	if (network->driver && network->driver->disconnect)
 		return network->driver->disconnect(network);
 
@@ -836,7 +844,10 @@ int connman_network_set_string(struct connman_network *network,
 {
 	DBG("network %p key %s value %s", network, key, value);
 
-	if (g_str_equal(key, "Name") == TRUE) {
+	if (g_str_equal(key, "Address") == TRUE) {
+		g_free(network->address);
+		network->address = g_strdup(value);
+	} else if (g_str_equal(key, "Name") == TRUE) {
 		g_free(network->name);
 		network->name = g_strdup(value);
 	} else if (g_str_equal(key, "Node") == TRUE) {
@@ -868,7 +879,9 @@ const char *connman_network_get_string(struct connman_network *network,
 {
 	DBG("network %p key %s", network, key);
 
-	if (g_str_equal(key, "Name") == TRUE)
+	if (g_str_equal(key, "Address") == TRUE)
+		return network->address;
+	else if (g_str_equal(key, "Name") == TRUE)
 		return network->name;
 	else if (g_str_equal(key, "Node") == TRUE)
 		return network->node;
@@ -1082,12 +1095,40 @@ static void network_remove(struct connman_element *element)
 		network->driver->remove(network);
 }
 
+static void network_change(struct connman_element *element)
+{
+	struct connman_network *network = element->network;
+
+	DBG("element %p name %s", element, element->name);
+
+	if (element->state != CONNMAN_ELEMENT_STATE_ERROR)
+		return;
+
+	if (element->error != CONNMAN_ELEMENT_ERROR_DHCP_FAILED)
+		return;
+
+	if (network->connected == FALSE)
+		return;
+
+	connman_element_unregister_children(element);
+
+	connman_device_set_disconnected(network->device, TRUE);
+
+	if (network->driver && network->driver->disconnect) {
+		network->driver->disconnect(network);
+		return;
+	}
+
+	network->connected = FALSE;
+}
+
 static struct connman_driver network_driver = {
 	.name		= "network",
 	.type		= CONNMAN_ELEMENT_TYPE_NETWORK,
 	.priority	= CONNMAN_DRIVER_PRIORITY_LOW,
 	.probe		= network_probe,
 	.remove		= network_remove,
+	.change		= network_change,
 };
 
 static int network_init(struct connman_device *device)
@@ -1103,6 +1144,7 @@ static int network_load(struct connman_network *network)
 	gchar *pathname, *data = NULL;
 	gsize length;
 	const char *name;
+	char *str;
 	int val;
 
 	DBG("network %p", network);
@@ -1140,13 +1182,19 @@ static int network_load(struct connman_network *network)
 	if (val > 0)
 		network->priority = val;
 
-	g_free(network->wifi.security);
-	network->wifi.security = g_key_file_get_string(keyfile,
+	str = g_key_file_get_string(keyfile,
 				network->identifier, "WiFi.Security", NULL);
+	if (str != NULL) {
+		g_free(network->wifi.security);
+		network->wifi.security = str;
+	}
 
-	g_free(network->wifi.passphrase);
-	network->wifi.passphrase = g_key_file_get_string(keyfile,
+	str = g_key_file_get_string(keyfile,
 				network->identifier, "WiFi.Passphrase", NULL);
+	if (str != NULL) {
+		g_free(network->wifi.passphrase);
+		network->wifi.passphrase = str;
+	}
 
 	g_key_file_free(keyfile);
 
