@@ -2,7 +2,7 @@
  *
  *  Connection Manager
  *
- *  Copyright (C) 2007-2008  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2007-2009  Intel Corporation. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -28,30 +28,53 @@
 
 #include "connman.h"
 
+int __connman_profile_add_device(struct connman_device *device)
+{
+	DBG("device %p", device);
+
+	return 0;
+}
+
+int __connman_profile_remove_device(struct connman_device *device)
+{
+	DBG("device %p", device);
+
+	return 0;
+}
+
+int __connman_profile_add_network(struct connman_network *network)
+{
+	DBG("network %p", network);
+
+	return 0;
+}
+
+int __connman_profile_remove_network(struct connman_network *network)
+{
+	DBG("network %p", network);
+
+	return 0;
+}
+
+const char *__connman_profile_active(void)
+{
+	DBG("");
+
+	return "/profile/default";
+}
+
 void __connman_profile_list(DBusMessageIter *iter)
 {
-	const char *path = "/profile/default";
+	const char *path = __connman_profile_active();
 
 	DBG("");
 
 	dbus_message_iter_append_basic(iter, DBUS_TYPE_OBJECT_PATH, &path);
 }
 
-static void append_string(DBusMessageIter *dict, const char *key, void *val)
+void __connman_profile_list_services(DBusMessageIter *iter)
 {
-	DBusMessageIter entry, value;
-
-	dbus_message_iter_open_container(dict, DBUS_TYPE_DICT_ENTRY,
-								NULL, &entry);
-
-	dbus_message_iter_append_basic(&entry, DBUS_TYPE_STRING, &key);
-
-	dbus_message_iter_open_container(&entry, DBUS_TYPE_VARIANT,
-					DBUS_TYPE_STRING_AS_STRING, &value);
-	dbus_message_iter_append_basic(&value, DBUS_TYPE_STRING, val);
-	dbus_message_iter_close_container(&entry, &value);
-
-	dbus_message_iter_close_container(dict, &entry);
+	DBG("");
 }
 
 static DBusMessage *get_properties(DBusConnection *conn,
@@ -74,7 +97,8 @@ static DBusMessage *get_properties(DBusConnection *conn,
 			DBUS_TYPE_STRING_AS_STRING DBUS_TYPE_VARIANT_AS_STRING
 			DBUS_DICT_ENTRY_END_CHAR_AS_STRING, &dict);
 
-	append_string(&dict, "Name", &name);
+	connman_dbus_dict_append_variant(&dict, "Name",
+						DBUS_TYPE_STRING, &name);
 
 	dbus_message_iter_close_container(&array, &dict);
 
